//! Stream buffering for sync of encrypted socked.
use super::*;
use smallvec::SmallVec;
use std::io;
use std::{
    task::{
	Poll, Context,
    },
    pin::Pin,
};
use bytes::{
    Buf, BufMut,
};

/// A wrapping buffer over a writer and/or reader.
#[pin_project]
#[derive(Debug)]
pub struct Buffered<T: ?Sized, const SIZE: usize>
{
    /// When an I/O error is hit when reading /writing into/from the internal buffer, store it here if there is data to dump from it.
    /// When there is no more data to return, return this error and then set `poisoned` to true.
    ///
    /// After that, all calls to `poll_read()`/`poll_write()` will return a new error indicating the stream has faulted.
    error: Option<io::Error>,
    /// If an error has been set above and/or released from the above slot, set this to true.
    /// Then all subsequent reads or writes will return a new error.
    poisoned: bool,

    /// Current internal buffer
    /// When it's full to `SIZE`, it should be written to `stream` at once then cleared when it's been written.
    buffer: SmallVec<[u8; SIZE]>, //TODO: Can we have a non-spilling stack vec?
    #[pin] stream: T
}

#[inline] fn div_mod<V>(a: V, b: V) -> (V, <V as std::ops::Div>::Output, <V as std::ops::Rem>::Output)
where V: std::ops::Div + std::ops::Rem + Clone
{
    (a.clone(), a.clone() / b.clone(), a % b)
}

impl<T: ?Sized, const SIZE: usize> AsyncWrite for Buffered<T, SIZE>
where T: AsyncWrite
{
    fn poll_write(self: Pin<&mut Self>, cx: &mut Context<'_>, buf: &[u8]) -> Poll<Result<usize, io::Error>> {

	let this = self.project();
	let w = if buf.len() + this.buffer.len() > SIZE {
	    // `buf` cannot fit into buffer, push the bytes that can fit.
	    let buf = &buf[..(this.buffer.len()-SIZE)];
	    this.buffer.extend_from_slice(buf);
	    buf.len()
	} else {
	    // `buf` can fit into buffer.
	    this.buffer.extend_from_slice(&buf[..]);
	    buf.len()
	};

	// Push `SIZE` bytes from buffer,
	loop {
	    if this.buffer.len() >= SIZE {
		use futures::ready;
		
		match ready!(this.stream.poll_write(cx, &this.buffer[..SIZE])) {
		    Ok(n) => {
			this.buffer.drain(0..n);
		    },
		    x => return Poll::Ready(x),
		    //TODO: How to break after this?
		}
	    } else {
		// TODO: How to re-queue for polling?
		todo!()
	    }
	}
	
	Poll::Ready(Ok(w))
	/*
	    match div_mod(buf.len(), SIZE) {
	// Buffer is equal to `SIZE`
	    (sz, _, _) if sz == SIZE => (),
	// Buffer is smaller than `SIZE`
	    (sz, 0, x) => (),
	// Buffer is a multiple of SIZE,
	    (sz, x, 0) => (),
    }*/
	// TODO: Push `buf` into `self.buffer`. When `self.buffer` is full, write to `self.stream` and clear what's been written.
    }
    fn poll_flush(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), io::Error>> {
	todo!("Flush any remaining bytes in the buffer to `stream`")
    }
    fn poll_shutdown(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), io::Error>> {
	// I think we're supposed to imply flush here?
	/*
	(&mut self).poll_flush(cx)?;
	let this = self.project();
	
	this.stream.poll_flush(cx)?;
	this.stream.poll_shutdown(cx)
	 */
	todo!("How to do the above?")
    }
}

